from input_algorithms.errors import BadSpec, BadSpecValue
from delfick_error import DelfickError, ProgrammerError

class AwsSyncrError(DelfickError): pass

# Explicitly make these errors in this context
BadSpec = BadSpec
BadSpecValue = BadSpecValue
ProgrammerError = ProgrammerError

class BadConfiguration(AwsSyncrError):
    desc = "Bad configuration"

class BadOptionFormat(AwsSyncrError):
    desc = "Bad option format"

class BadOption(AwsSyncrError):
    desc = "Bad Option"

class BadYaml(AwsSyncrError):
    desc = "Invalid yaml file"

class UserQuit(AwsSyncrError):
    desc = "User quit the program"

class BadEnvironment(AwsSyncrError):
    desc = "Something bad in the environment"

class BadTask(AwsSyncrError):
    desc = "Bad task"

class BadCredentials(AwsSyncrError):
    desc = "Bad credentials"

class BadPolicy(AwsSyncrError):
    desc = "Bad policy"

class InvalidDocument(AwsSyncrError):
    desc = "Bad document"

class BadAmazon(AwsSyncrError):
    desc = "Bad Amazon"

class BadTemplate(AwsSyncrError):
    desc = "Bad template"

class InvalidGrant(AwsSyncrError):
    desc = "Bad Grant"

class BadImport(AwsSyncrError):
    desc = "Failed to import"

class UnknownStage(AwsSyncrError):
    desc = "Unknown stage"

class UnsyncedGateway(AwsSyncrError):
    desc = "Gateway not synced"

class UnknownZone(AwsSyncrError):
    desc = "Unknown zone"

class UnknownEndpoint(AwsSyncrError):
    desc = "Unknown endpoint"




from aws_syncr.amazon.common import AmazonMixin
from aws_syncr.errors import AwsSyncrError
from aws_syncr.differ import Differ

from input_algorithms.spec_base import NotSpecified
import requests
import boto3

import logging
import json
import six
import re

log = logging.getLogger("aws_syncr.amazon.apigateway")

class ApiGateway(AmazonMixin, object):
    def __init__(self, amazon, environment, accounts, dry_run):
        self.amazon = amazon
        self.dry_run = dry_run

        self.accounts = accounts
        self.account_id = accounts[environment]
        self.environment = environment

        self.client = lambda region: self.amazon.session.client('apigateway', region)

    def gateway_info(self, gateway_name, region):
        client = self.client(region)
        apis = client.get_rest_apis()

        for item in apis['items']:
            if item['name'] == gateway_name:
                identity = item['id']
                info = {"identity": identity, 'name': gateway_name}
                self.load_info(client, info)
                return info

    def load_info(self, client, info):
        """Fill out information about the gateway"""
        if 'identity' in info:
            info['stages'] = client.get_stages(restApiId=info['identity'])['item']
            info['resources'] = client.get_resources(restApiId=info['identity'])['items']
            for resource in info['resources']:
                for method in resource.get('resourceMethods', {}):
                    resource['resourceMethods'][method] = client.get_method(restApiId=info['identity'], resourceId=resource['id'], httpMethod=method)
                    for status_code, options in resource['resourceMethods'][method]['methodResponses'].items():
                        options.update(client.get_method_response(restApiId=info['identity'], resourceId=resource['id'], httpMethod=method, statusCode=status_code))

            info['deployment'] = client.get_deployments(restApiId=info['identity'])['items']
        else:
            for key in ('stages', 'resources', 'deployment'):
                info[key] = []

        info['api_keys'] = client.get_api_keys()['items']
        info['domains'] = client.get_domain_names()['items']
        for domain in info['domains']:
            domain['mappings'] = client.get_base_path_mappings(domainName=domain['domainName']).get('items', [])

    def create_gateway(self, name, location, stages, resources, api_keys, domains):
        client = self.client(location)

        with self.catch_boto_400("Couldn't Make gateway", gateway=name):
            for _ in self.change("+", "gateway", gateway=name):
                result = client.create_rest_api(name=name)
                info = {"identity": result['id']}
                self.load_info(client, info)
                self.modify_gateway(info, name, location, stages, resources, api_keys, domains)

        if self.dry_run:
            info = {}
            self.load_info(client, info)
            self.modify_gateway(info, name, location, stages, resources, api_keys, domains)

    def modify_gateway(self, gateway_info, name, location, stages, resources, api_keys, domains):
        client = self.client(location)

        current_domain_names = [domain['domainName'] for domain in gateway_info['domains']]
        missing = set(d.full_name for d in domains.values()) - set(current_domain_names)

        for domain in missing:
            with self.catch_boto_400("Couldn't Make domain", domain=domain):
                for _ in self.change("+", "domain", domain=domain):
                    certificate = [d for d in domains.values() if d.full_name == domain][0].certificate
                    client.create_domain_name(domainName=domain
                        , certificateName = certificate.name
                        , certificateBody = certificate.body.resolve(self.amazon)
                        , certificateChain = certificate.chain.resolve(self.amazon)
                        , certificatePrivateKey = certificate.key.resolve(self.amazon)
                        )

        self.modify_resources(client, gateway_info, location, name, resources)
        self.modify_stages(client, gateway_info, name, stages)

        self.modify_domains(client, gateway_info, name, domains)
        self.modify_api_keys(client, gateway_info, name, api_keys)

    def modify_resources(self, client, gateway_info, location, name, resources):
        current_resources = [r['path'] for r in gateway_info['resources']]
        wanted_resources = [r for r in resources]

        resources_by_path = dict((r['path'], r) for r in gateway_info['resources'])

        for_removal = [key for key in list(set(current_resources) - set(wanted_resources)) if key != '/']
        for_addition = [key for key in list(set(wanted_resources) - set(current_resources)) if key != '/']
        for_modification = list(set(['/'] + [r for r in wanted_resources if r in current_resources] + list(for_addition)))

        for path in for_removal:
            with self.catch_boto_400("Couldn't remove resource", gateway=name, resource=path):
                for _ in self.change("-", "gateway resource", gateway=name, resource=path):
                    resource_id = resources_by_path[path]['id']
                    client.delete_resource(restApiId=gateway_info['identity'], resourceId=resource_id)

        for path in for_addition:
            with self.catch_boto_400("Couldn't add resource", gateway=name, resource=path):
                for _ in self.change("+", "gateway resource", gateway=name, resource=path):
                    parent_id = resources_by_path['/']['id']
                    while path and path.startswith("/"):
                        path = path[1:]

                    upto = ['']
                    for part in path.split('/'):
                        upto.append(part)
                        if '/'.join(upto) not in resources_by_path:
                            info = client.create_resource(restApiId=gateway_info['identity'], parentId=parent_id, pathPart=part)
                            resources_by_path['/'.join(upto)] = info
                            parent_id = info['id']
                        else:
                            parent_id = resources_by_path['/'.join(upto)]['id']

        for path in for_modification:
            wanted_methods = {}
            if path in resources:
                wanted_methods = dict(resources[path].method_options)

            current_methods = resources_by_path.get(path, {}).get('resourceMethods', {})
            self.modify_resource_methods(client, gateway_info, location, name, path, current_methods, wanted_methods, resources_by_path)

    def modify_resource_methods(self, client, gateway_info, location, name, path, old_methods, new_methods, resources_by_path):
        for_removal = set(old_methods) - set(new_methods)
        for_addition = set(new_methods) - set(old_methods)
        for_modification = [method for method in new_methods if method in old_methods] + list(for_addition)

        for method in for_removal:
            with self.catch_boto_400("Couldn't remove method", gateway=name, resource=path, method=method):
                for _ in self.change("-", "gateway resource method", gateway=name, resource=path, method=method):
                    resource_id = resources_by_path[path]['id']
                    client.delete_method(restApiId=gateway_info['identity'], resourceId=resource_id, httpMethod=method)

        for method in for_addition:
            with self.catch_boto_400("Couldn't add method", gateway=name, resource=path, method=method):
                for _ in self.change("+", "gateway resource method", gateway=name, resource=path, method=method):
                    resource_id = resources_by_path[path]['id']
                    client.put_method(restApiId=gateway_info['identity'], resourceId=resource_id, httpMethod=method
                        , apiKeyRequired=new_methods[method].method_request.require_api_key
                        , authorizationType = "none"
                        )

        for method in for_modification:
            with self.catch_boto_400("Couldn't modify method", gateway=name, resource=path, method=method):
                if method in old_methods:
                    if old_methods[method]['apiKeyRequired'] != new_methods[method].method_request.require_api_key:
                        for _ in self.change("M", "gateway resource method", gateway=name, resource=path, method=method):
                            resource_id = resources_by_path[path]['id']
                            operations = [{"op": "replace", "path": "/apiKeyRequired", "value": str(new_methods[method].method_request.require_api_key)}]
                            client.update_method(restApiId=gateway_info['identity'], resourceId=resource_id, httpMethod=method, patchOperations=operations)

                self.modify_resource_method_status_codes(client, gateway_info, name, path, method, old_methods.get(method, {}), new_methods[method], resources_by_path)
                self.modify_resource_method_integration(client, gateway_info, location, name, path, method, old_methods.get(method, {}), new_methods[method], resources_by_path)
                self.modify_resource_method_integration_response(client, gateway_info, name, path, method, old_methods.get(method, {}), new_methods[method], resources_by_path)

    def modify_resource_method_status_codes(self, client, gateway_info, name, path, method, old_method, new_method, resources_by_path):
        old_status_codes = list(old_method.get('methodResponses', {}).keys())
        new_status_codes = list(str(st) for st in new_method.method_response.responses.keys())

        for_removal = set(old_status_codes) - set(new_status_codes)
        for_addition = set(new_status_codes) - set(old_status_codes)
        for_modification = [status_code for status_code in new_status_codes if status_code in old_status_codes]

        for status_code in for_removal:
            for _ in self.change("-", "gateway resource method response", gateway=name, resource=path, method=method, status_code=status_code):
                resource_id = resources_by_path[path]['id']
                client.delete_method_response(restApiId=gateway_info['identity'], resourceId=resource_id, httpMethod=method, statusCode=str(status_code))

        for status_code in for_addition:
            for _ in self.change("+", "gateway resource method response", gateway=name, resource=path, method=method, status_code=status_code):
                resource_id = resources_by_path[path]['id']
                models = {new_method.method_response.responses[int(status_code)]: "Empty"}
                models = dict((ct, model) for ct, model in models.items() if ct != "application/json")
                client.put_method_response(restApiId=gateway_info['identity'], resourceId=resource_id, httpMethod=method, statusCode=str(status_code)
                    , responseParameters = {}
                    , responseModels = models
                    )

        for status_code in for_modification:
            new = {new_method.method_response.responses[int(status_code)]: "Empty"}
            new = dict((ct, model) for ct, model in new.items() if ct != "application/json")

            old = old_method["methodResponses"][status_code].get("responseModels", {})
            changes = list(Differ.compare_two_documents(old, new))

            old = dict((ct.replace('/', '~1'), v) for ct, v in old.items())
            new = dict((ct.replace('/', '~1'), v) for ct, v in new.items())

            if changes:
                for_removal = set(old) - set(new)
                for_addition = set(new) - set(old)
                for_mod = [content_type for content_type in new if content_type in old]
                operations = []

                for content_type in for_removal:
                    operations.append({"op": "remove", "path": "/responseModels/{0}".format(content_type)})
                for content_type in for_addition:
                    operations.append({"op": "add", "path":"/responseModels/{0}".format(content_type), 'value': new[content_type]})
                for content_type in for_mod:
                    operations.append({"op": "replace", "path":"/responseModels/{0}".format(content_type), 'value': new[content_type]})

                for _ in self.change("M", "gateway resource method response model", gateway=name, resource=path, method=method, status_code=status_code, changes=changes):
                    resource_id = resources_by_path[path]['id']
                    client.update_method_response(restApiId=gateway_info['identity'], resourceId=resource_id, httpMethod=method, statusCode=str(status_code)
                        , patchOperations = operations
                        )

    def modify_resource_method_integration(self, client, gateway_info, location, name, path, method, old_method, new_method, resources_by_path):
        old_integration = old_method.get('methodIntegration', {})
        new_integration = new_method.integration_request

        new_kwargs = new_integration.put_kwargs(location, self.accounts, self.environment)
        old_kwargs = {} if not old_integration else {"type": old_integration["type"], "httpMethod": old_integration.get("httpMethod")}

        if old_integration and old_integration.get('requestTemplates'):
            old_kwargs['requestTemplates'] = old_integration['requestTemplates']
            for ct, template in list(old_kwargs['requestTemplates'].items()):
                if not template:
                    old_kwargs['requestTemplates'][ct] = ""

        if old_kwargs and old_kwargs['type'] == 'AWS':
            old_kwargs['uri'] = old_integration['uri']
        elif old_kwargs and old_kwargs['type'] == 'MOCK':
            old_kwargs["httpMethod"] = method

        changes = list(Differ.compare_two_documents(old_kwargs, new_kwargs))

        # Make sure our integration can be called by apigateway
        if 'identity' in gateway_info:
            arn = "arn:aws:execute-api:{0}:{1}:{2}/*/".format(location, self.account_id, gateway_info['identity'])
            new_integration.create_permissions(self.amazon, arn, name, self.accounts, self.environment)
        else:
            # Only possible in dry-run
            new_integration.announce_create_permissions(name, self.change)

        if changes:
            symbol = "+" if not old_integration else 'M'
            for _ in self.change(symbol, "gateway resource method integration request", gateway=name, resource=path, method=method, type=new_kwargs['type'], changes=changes):
                resource_id = resources_by_path[path]['id']
                integration_method = new_kwargs.pop("httpMethod")
                res = client.put_integration(restApiId=gateway_info['identity'], resourceId=resource_id, httpMethod=method
                    , integrationHttpMethod=integration_method
                    , **new_kwargs
                    )

                # put_integration removes the integration response so we take it away from our record
                # And let modify_resource_method_integration_response deal with the dissapearance
                if 'integrationResponses' in old_method.get("methodIntegration", {}):
                    del old_method["methodIntegration"]['integrationResponses']
                new_kwargs['responseTemplates'] = new_method.integration_response.responses.items()

    def modify_resource_method_integration_response(self, client, gateway_info, name, path, method, old_method, new_method, resources_by_path):
        old_integration = old_method.get('methodIntegration', {}).get("integrationResponses", {})
        wanted_integration = dict((str(s), v) for s, v in new_method.integration_response.responses.items())

        for_removal = set(old_integration) - set(wanted_integration)
        for_addition = set(wanted_integration) - set(old_integration)
        for_modification = [s for s in wanted_integration if s in old_integration]

        for status_code in for_removal:
            for _ in self.change("-", "gateway resource integration response", gateway=name, resource=path, method=method, status_code=status_code):
                resource_id = resources_by_path[path]['id']
                client.delete_integration_response(restApiId=gateway_info['identity'], resourceId=resource_id, httpMethod=method, statusCode=str(status_code))

        for status_code in for_addition:
            for _ in self.change("+", "gateway resource integration response", gateway=name, resource=path, method=method, status_code=status_code):
                resource_id = resources_by_path[path]['id']
                client.put_integration_response(restApiId=gateway_info['identity'], resourceId=resource_id, httpMethod=method, statusCode=str(status_code)
                    , responseTemplates = {} if not wanted_integration[status_code] else dict((m.content_type, m.template) for m in wanted_integration[status_code])
                    )

        # Modify response Templates
        for status_code in for_modification:
            old = old_integration[status_code].get('responseTemplates', {})
            for ct, template in old.items():
                if template is None:
                    old[ct] = ""

            new = {}
            if wanted_integration[status_code]:
                new = dict((m.content_type, m.template) for m in wanted_integration[status_code])
            changes = list(Differ.compare_two_documents(old, new))

            old = dict((ct.replace('/', '~1'), v) for ct, v in old.items())
            new = dict((ct.replace('/', '~1'), v) for ct, v in new.items())

            if changes:
                for_removal = set(old) - set(new)
                for_addition = set(new) - set(old)
                for_mod = [content_type for content_type in new if content_type in old]
                operations = []

                for content_type in for_removal:
                    operations.append({"op": "remove", "path": "/responseTemplates/{0}".format(content_type)})
                for content_type in for_addition:
                    operations.append({"op": "add", "path":"/responseTemplates/{0}".format(content_type), 'value': new[content_type]})
                for content_type in for_mod:
                    operations.append({"op": "replace", "path":"/responseTemplates/{0}".format(content_type), 'value': new[content_type]})

                for _ in self.change("M", "gateway resource integration response", gateway=name, resource=path, method=method, status_code=status_code, changes=changes):
                    resource_id = resources_by_path[path]['id']
                    client.update_integration_response(restApiId=gateway_info['identity'], resourceId=resource_id, httpMethod=method, statusCode=str(status_code)
                        , patchOperations = operations
                        )

    def modify_stages(self, client, gateway_info, name, stages):
        current_stages = [stage['stageName'] for stage in gateway_info['stages']]
        missing = set(stages) - set(current_stages)
        for_removal = set(current_stages) - set(stages)

        for stage in for_removal:
            with self.catch_boto_400("Couldn't remove stage", gateway=name, stage=stage):
                for _ in self.change("-", "gateway stage", gateway=name, stage=stage):
                    client.delete_stage(restApiId=gateway_info['identity'], stageName=stage)

        if 'identity' in gateway_info:
            deployments = client.get_deployments(restApiId=gateway_info['identity'])['items']
            stages = client.get_stages(restApiId=gateway_info['identity'])['item']
        else:
            stages = []
            deployments = []

        stage_deployments = [stage['deploymentId'] for stage in stages]
        for_removal = [deployment['id'] for deployment in deployments if deployment['id'] not in stage_deployments]
        for deployment in for_removal:
            with self.catch_boto_400("Couldn't remove deployment", gateway=name, deployment=deployment):
                for _ in self.change("-", "gateway deployment", gateway=name, deployment=deployment):
                    client.delete_deployment(restApiId=gateway_info['identity'], deploymentId=deployment)

        for stage in missing:
            with self.catch_boto_400("Couldn't add stage", gateway=name, stage=stage):
                for _ in self.change("+", "gateway stage", gateway=name, stage=stage):
                    if gateway_info.get('deployment'):
                        client.create_stage(restApiId=gateway_info['identity'], deploymentId=gateway_info['deployment'][0]['id'], stageName=stage)
                    else:
                        client.create_deployment(restApiId=gateway_info['identity'], stageName=stage)

    def modify_api_keys(self, client, gateway_info, name, api_keys):
        current = [ak['name'] for ak in gateway_info['api_keys']]
        wanted = [api_key.name for api_key in api_keys]

        for_addition = list(set(wanted) - set(current))
        for keyname in for_addition:
            with self.catch_boto_400("Couldn't add api keys", api_key=keyname):
                for _ in self.change("+", "gateway api key", gateway=name, api_key=keyname):
                    api_key = [api_key for api_key in api_keys if api_key.name == keyname][0]
                    client.create_api_key(name=keyname, enabled=True
                        , stageKeys=[{'restApiId': gateway_info['identity'], 'stageName': stage} for stage in api_key.stages]
                        )

        for_modification = [key for key in current if key in wanted]
        for keyname in for_modification:
            with self.catch_boto_400("Couldn't modify api keys", api_key=keyname):
                api_key = [api_key for api_key in api_keys if api_key.name == keyname][0]
                old_api_key = [ak for ak in gateway_info['api_keys'] if api_key['name'] == keyname][0]
                other_api_stages = [key for key in old_api_key['stageKeys'] if key[:key.find('/')] != gateway_info.get('identity')]

                operations = []

                new_stage_keys = ["{0}/{1}".format(gateway_info.get('identity'), key) for key in api_key.stages] + other_api_stages
                changes = list(Differ.compare_two_documents(sorted(old_api_key['stageKeys']), sorted(new_stage_keys)))

                if changes:
                    for_removal = set(old_api_key['stageKeys']) - set(new_stage_keys)
                    for key in for_removal:
                        operations.append({"op": "remove", "path": "/stages", "value": key})

                    for_addition = set(new_stage_keys) - set(old_api_key['stageKeys'])
                    for key in for_addition:
                        operations.append({"op": "add", "path": "/stages", "value": key})

                if operations:
                    for _ in self.change("M", "gateway api key", gateway=name, api_key=keyname, changes=changes):
                        client.update_api_key(apiKey=old_api_key['id'], patchOperations=operations)

    def modify_domains(self, client, gateway_info, name, domains):
        for domain in domains.values():
            found = []
            matches = [d for d in gateway_info['domains'] if d['domainName'] == domain.full_name]
            if matches:
                for mapping in matches[0]['mappings']:
                    if ('identity' in gateway_info and mapping['restApiId'] == gateway_info['identity']) or mapping['basePath'] == domain.base_path:
                        found.append(mapping)

            current = [dict((key, mapping.get(key)) for key in ('restApiId', 'stage', 'basePath')) for mapping in found]
            wanted = {'restApiId': gateway_info.get('identity', '<gateway id>'), 'stage': domain.stage, 'basePath': domain.base_path}

            if list(Differ.compare_two_documents(current, [wanted])):
                for_removal = [mapping for mapping in current if mapping['basePath'] != wanted['basePath']]
                for_addition = [mapping for mapping in [wanted] if mapping['basePath'] not in [m['basePath'] for m in found]]

                for_modification = []
                for new in [wanted]:
                    for old in found:
                        if old['basePath'] == new['basePath']:
                            for_modification.append((old, new))

                with self.catch_boto_400("Couldn't remove domain name bindings", gateway=name):
                    for mapping in for_removal:
                        for _ in self.change("-", "domain name gateway association", gateway=name, base_path=mapping['basePath']):
                            client.update_base_path_mapping(domainName=domain.full_name, basePath=mapping['basePath']
                                , patchOperations = [{"op": "remove", "path": "/"}]
                                )

                with self.catch_boto_400("Couldn't add domain name bindings", gateway=name):
                    for mapping in for_addition:
                        for _ in self.change("+", "domain name gateway association", gateway=name, base_path=mapping['basePath'], stage=mapping['stage']):
                            client.create_base_path_mapping(domainName=domain.full_name, basePath=mapping['basePath'], restApiId=gateway_info['identity'], stage=mapping['stage'])

                with self.catch_boto_400("Couldn't modify domain name bindings", gateway=name):
                    for old, new in for_modification:
                        changes = Differ.compare_two_documents(old, new)
                        for _ in self.change("M", "domain name gateway association", gateway=name, stage=new['stage'], base_path=new["basePath"], changes=changes):
                            operations = []

                            if old['restApiId'] != new['restApiId']:
                                operations.append({"op": "replace", "path": "/restapiId", "value": new['restApiId']})

                            if old.get('stage') != new.get('stage'):
                                operations.append({"op": "replace", "path": "/stage", "value": new['restApiId']})

                            client.update_base_path_mapping(domainName=domain.full_name, basePath=wanted['basePath'], patchOperations = operations)

    def deploy_stage(self, gateway_info, location, stage, description):
        client = self.client(location)
        for _ in self.change("D", "Deployment", gateway=gateway_info['name'], stage=stage):
            log.info("Deploying stage {0} for gateway {1}".format(stage, gateway_info['name']))
            client.create_deployment(restApiId=gateway_info['identity'], stageName=stage, description=description)
            print("https://{0}.execute-api.{1}.amazonaws.com/{2}".format(gateway_info['identity'], location, stage))

        previous_deployments = [s['deploymentId'] for s in gateway_info['stages'] if s['stageName'] == stage]
        if previous_deployments:
            for previous_deployment in previous_deployments:
                for _ in self.change("-", "deployment", gateway=gateway_info['name'], deployment=previous_deployment):
                    client.delete_deployment(restApiId=gateway_info['identity'], deploymentId=previous_deployment)

    def cname_for(self, gateway_location, record):
        with self.ignore_missing():
            return self.client(gateway_location).get_domain_name(domainName=record)['distributionDomainName']
        raise AwsSyncrError("Please do a sync first!")

    def test_stage(self, gateway_info, location, stage, method, endpoint, sample_event, desired_output_for_test):
        kwargs = {}
        if sample_event:
            if not isinstance(sample_event, six.string_types) and sample_event is not NotSpecified:
                kwargs['data'] = json.dumps(dict(sample_event.items()))

        # Find the url to use
        url = "https://{0}.execute-api.{1}.amazonaws.com/{2}".format(gateway_info['identity'], location, stage)
        for domain in gateway_info['domains']:
            if 'mappings' in domain:
                for mapping in domain['mappings']:
                    if mapping['restApiId'] == gateway_info['identity'] and mapping['stage'] == stage:
                        url = "https://{0}".format(domain['domainName'])
        url = "{0}{1}".format(url, endpoint)
        log.info("{0}ing to {1}".format(method, url))

        # Find an api-key
        api_key = ""
        stagekey = "{0}/{1}".format(gateway_info['identity'], stage)
        for api_key in gateway_info['api_keys']:
            if stagekey in api_key['stageKeys']:
                log.info("Found an api key to use ({0})".format(api_key['name']))
                api_key = api_key['id']
                break

        # Use the api key if it is required
        resource = [r for r in gateway_info['resources'] if r['path'] == endpoint][0]
        method_options = [o for m, o in resource['resourceMethods'].items() if m == method][0]
        if method_options['apiKeyRequired']:
            kwargs['headers'] = {'x-api-key': api_key}

        # make the request
        res = getattr(requests, method.lower())(url, **kwargs)

        print("Got result with status_code={0}".format(res.status_code))
        print(res.content.decode('utf-8'))

        if res.status_code != 200:
            # Say we failed if status code isn't 200
            return False
        else:
            # Say we succeeded if we meet the desired_output_for_test
            if desired_output_for_test and desired_output_for_test is not NotSpecified:
                if isinstance(desired_output_for_test, six.string_types):
                    content = res.content.decode('utf-8')
                    if not re.match(desired_output_for_test, content):
                        print("content '{0}' does not match pattern '{1}'".format(content, desired_output_for_test))
                        return False

                else:
                    content = json.loads(res.content.decode('utf-8'))
                    if any(key not in content or content[key] != val for key, val in desired_output_for_test.items()):
                        print("Not all of the values match our desired output of '{0}'".format(desired_output_for_test))
                        return False

        return True




from datadiff import diff
import logging
import json
import six

log = logging.getLogger("aws_syncr.operations.differ")

class Differ(object):
    @classmethod
    def compare_two_documents(kls, doc1, doc2):
        """Compare two documents by converting them into json objects and back to strings and compare"""
        first = doc1
        if isinstance(doc1, six.string_types):
            try:
                first = json.loads(doc1)
            except (ValueError, TypeError) as error:
                log.warning("Failed to convert doc into a json object\terror=%s", error)
                yield error.args[0]
                return

        second = doc2
        if isinstance(doc2, six.string_types):
            try:
                second = json.loads(doc2)
            except (ValueError, TypeError) as error:
                log.warning("Failed to convert doc into a json object\terror=%s", error)
                yield error.args[0]
                return

        # Ordering the principals because the ordering amazon gives me hates me
        def sort_statement(statement):
            for principal in (statement.get("Principal", None), statement.get("NotPrincipal", None)):
                if principal:
                    for principal_type in ("AWS", "Federated", "Service"):
                        if principal_type in principal and type(principal[principal_type]) is list:
                            principal[principal_type] = sorted(principal[principal_type])
        def sort_key(statement, key):
            if key in statement and type(statement[key]) is list:
                statement[key] = sorted(statement[key])
        for document in (first, second):
            if "Statement" in document:
                if type(document["Statement"]) is dict:
                    sort_statement(document["Statement"])
                    sort_key(document["Statement"], "Action")
                    sort_key(document["Statement"], "NotAction")
                    sort_key(document["Statement"], "Resource")
                    sort_key(document["Statement"], "NotResource")
                else:
                    for statement in document["Statement"]:
                        sort_statement(statement)
                        sort_key(statement, "Action")
                        sort_key(statement, "NotAction")
                        sort_key(statement, "Resource")
                        sort_key(statement, "NotResource")

        difference = diff(first, second, fromfile="current", tofile="new").stringify()
        if difference:
            lines = difference.split('\n')
            if not first or not second or first != second:
                for line in lines:
                    yield line




from aws_syncr.amazon.common import AmazonMixin
from aws_syncr.errors import UnknownZone
from aws_syncr.differ import Differ

import boto3

import logging

log = logging.getLogger("aws_syncr.amazon.route53")

class Route53(AmazonMixin, object):
    def __init__(self, amazon, environment, accounts, dry_run):
        self.amazon = amazon
        self.dry_run = dry_run

        self.accounts = accounts
        self.account_id = accounts[environment]
        self.environment = environment

        self.client = self.amazon.session.client('route53')

    def route_info(self, route_name, zone):
        zones = self.client.list_hosted_zones_by_name(DNSName=zone)
        if not zones.get("HostedZones"):
            raise UnknownZone(zone=zone)

        info = {"zoneid": zones['HostedZones'][0]['Id'], 'zone': zone}

        next_record_name = ""
        next_record_type = ""
        next_record_identifier = ""
        while True:
            next_values = (("StartRecordName", next_record_name), ("StartRecordType", next_record_type), ("StartRecordIdentifier", next_record_identifier))
            kwargs = dict((name, val) for name, val in next_values if val)
            record_infos = self.client.list_resource_record_sets(HostedZoneId=info['zoneid'], **kwargs)

            for i in record_infos['ResourceRecordSets']:
                if i['Name'] == "{0}.{1}".format(route_name, zone):
                    info['record'] = i
                    return info

            if record_infos['IsTruncated']:
                next_record_name = record_infos['NextRecordName']
                next_record_type = record_infos['NextRecordType']
                next_record_identitifer = record_infos['NextRecordIdentifier']
            else:
                break

        # Didn't find the record
        return {}

    def create_route(self, name, zone, record_type, record_target):
        old = {}
        new = {"target": [{"Value": record_target}], 'type': record_type}
        changes = list(Differ.compare_two_documents(old, new))
        hosted_zone_id = self.client.list_hosted_zones_by_name(DNSName=zone)['HostedZones'][0]['Id']

        with self.catch_boto_400("Couldn't add record", record=name, zone=zone):
            for _ in self.change("+", "record", record=name, zone=zone, changes=changes):
                self.client.change_resource_record_sets(HostedZoneId=hosted_zone_id
                    , ChangeBatch = {"Changes": [
                          { "Action": "CREATE"
                          , "ResourceRecordSet":
                            { "Name": "{0}.{1}".format(name, zone)
                            , "Type": record_type
                            , "TTL": 60
                            , "ResourceRecords": new['target']
                            }
                          }
                        ]
                      }
                    )

    def modify_route(self, route_info, name, zone, record_type, record_target):
        old = {"target": route_info['record']['ResourceRecords'], "type": route_info['record']['Type']}
        new = {"target": [{"Value": record_target}], 'type': record_type}
        changes = list(Differ.compare_two_documents(old, new))
        hosted_zone_id = self.client.list_hosted_zones_by_name(DNSName=zone)['HostedZones'][0]['Id']

        if changes:
            with self.catch_boto_400("Couldn't change record", record=name, zone=zone):
                for _ in self.change("M", "record", record=name, zone=zone, changes=changes):
                    self.client.change_resource_record_sets(HostedZoneId=hosted_zone_id
                        , ChangeBatch = {"Changes": [
                              { "Action": "UPSERT"
                              , "ResourceRecordSet":
                                { "Name": "{0}.{1}".format(name, zone)
                                , "Type": record_type
                                , "TTL": 60
                                , "ResourceRecords": new['target']
                                }
                              }
                            ]
                          }
                        )




from aws_syncr.errors import BadAmazon, BadCredentials

from botocore.exceptions import ClientError, NoCredentialsError

from contextlib import contextmanager

class AmazonMixin:
    @contextmanager
    def catch_boto_400(self, message, heading=None, document=None, **info):
        """Turn a BotoServerError 400 into a BadAmazon"""
        try:
            yield
        except ClientError as error:
            if str(error.response["ResponseMetadata"]["HTTPStatusCode"]).startswith("4"):
                if heading or document:
                    print("=" * 80)
                    if heading:
                        print(heading)
                    print(document)
                    print("=" * 80)
                error_message = error.response["Error"]["Message"]
                raise BadAmazon(message, error_message=error_message, error_code=error.response["ResponseMetadata"]["HTTPStatusCode"], **info)
            else:
                raise

    @contextmanager
    def ignore_missing(self):
        try:
            yield
        except ClientError as error:
            if "ResponseMetadata" in error.response and error.response["ResponseMetadata"]["HTTPStatusCode"] == 404 or error.response["Error"]["Code"] == "404" or error.response["Error"]["Code"] == "NotFoundException":
                pass
            else:
                raise

    @contextmanager
    def catch_invalid_credentials(self):
        try:
            yield
        except NoCredentialsError:
            raise BadCredentials("Failed to find valid credentials")
        except ClientError as error:
            if error.response["ResponseMetadata"]["HTTPStatsuCode"] == 403:
                raise BadCredentials("Failed to find valid credentials", error=error.message)
            else:
                raise

    def print_change(self, symbol, typ, changes=None, document=None, **kwargs):
        """Print out a change"""
        values = ", ".join("{0}={1}".format(key, val) for key, val in sorted(kwargs.items()))
        print("{0} {1}({2})".format(symbol, typ, values))
        if changes:
            for change in changes:
                print("\n".join("\t{0}".format(line) for line in change.split('\n')))
        elif document:
            print("\n".join("\t{0}".format(line) for line in document.split('\n')))

    def change(self, symbol, typ, **kwargs):
        """Print out a change and then do the change if not doing a dry run"""
        self.print_change(symbol, typ, **kwargs)
        if not self.dry_run:
            try:
                yield
            except:
                raise
            else:
                self.amazon.changes = True




from aws_syncr.errors import InvalidDocument

from input_algorithms.dictobj import dictobj

import json

class Document(dictobj):
    fields = ["statements"]

    @property
    def document(self):
        if not self.statements:
            return None

        document = {"Version": "2012-10-17", "Statement": [s.statement for s in self.statements]}

        try:
            return json.dumps(document, indent=2)
        except (TypeError, ValueError) as err:
            raise InvalidDocument("Document wasn't valid json", error=err, document=document)




from aws_syncr.errors import BadTemplate, UnknownStage, UnsyncedGateway, UnknownEndpoint, AwsSyncrError
from aws_syncr.formatter import MergedOptionStringFormatter
from aws_syncr.option_spec.lambdas import Lambda

from Crypto.Util import Counter
from Crypto.Cipher import AES

from input_algorithms.spec_base import NotSpecified
from input_algorithms.validators import Validator
from input_algorithms.errors import BadSpecValue
from input_algorithms import spec_base as sb
from input_algorithms.spec_base import Spec
from input_algorithms.dictobj import dictobj

from option_merge import MergedOptions
from textwrap import dedent
import logging
import base64
import json
import sys
import six

log = logging.getLogger("aws_syncr.option_spec.apigateway")

formatted_string = lambda: sb.formatted(sb.string_or_int_as_string_spec(), MergedOptionStringFormatter)

api_key_spec = lambda: sb.create_spec(ApiKey
    , name = formatted_string()
    , stages = sb.listof(formatted_string())
    )

class formatted_dictionary(sb.Spec):
    def normalise(self, meta, val):
        val = sb.dictionary_spec().normalise(meta, val)
        return self.formatted_dict(meta, val)

    def formatted_dict(self, meta, val, chain=None):
        result = {}
        for key, val in val.items():
            if type(val) is dict:
                result[key] = self.formatted_dict(meta.at(key), val, chain)
            elif isinstance(val, six.string_types):
                result[key] = sb.formatted(sb.string_spec(), formatter=MergedOptionStringFormatter).normalise(meta.at(key), val)
            else:
                result[key] = val
        return result

class valid_secret(Validator):
    def validate(self, meta, val):
        val = sb.dictionary_spec().normalise(meta, val)
        if 'plain' in val and 'kms' in val:
            raise BadSpecValue("Please only specify plain or kms", got=list(val.keys()), meta=meta)

        if 'plain' not in val and 'kms' not in val:
            raise BadSpecValue("Please specify plain or kms", got=list(val.keys()), meta=meta)

        if 'kms' in val and ('location' not in val or 'kms_data_key' not in val):
            raise BadSpecValue("Please specify location and kms_data_key if you specify kms", got=list(val.keys()), meta=meta)

        return val

secret_spec = lambda: sb.create_spec(Secret
    , valid_secret()

    , plain = sb.optional_spec(formatted_string())
    , kms = sb.optional_spec(formatted_string())
    , location = sb.optional_spec(formatted_string())
    , kms_data_key = sb.optional_spec(formatted_string())
    )

class certificate_spec(Spec):
    def normalise(self, meta, val):
        if isinstance(val, six.string_types):
            val = formatted_string().normalise(meta, val)

        return sb.create_spec(Certificate
            , name = sb.required(formatted_string())
            , body = sb.required(secret_spec())
            , key = sb.required(secret_spec())
            , chain = sb.required(secret_spec())
            ).normalise(meta, val)

class custom_domain_name_spec(Spec):
    def setup(self, gateway_location):
        self.gateway_location = gateway_location

    def normalise(self, meta, val):
        name = meta.key_names()["_key_name_0"]
        result = sb.create_spec(DomainName
            , name = sb.overridden(name)
            , gateway_location = sb.overridden(self.gateway_location)
            , zone = formatted_string()
            , stage = formatted_string()
            , base_path = sb.defaulted(formatted_string(), "(none)")
            , certificate = sb.required(certificate_spec())
            ).normalise(meta, val)

        while result.zone and result.zone.endswith("."):
            result.zone = result.zone[:-1]

        return result

formatted_dictionary_or_string = lambda : sb.match_spec(
      (six.string_types, formatted_string())
    , fallback = sb.dictof(sb.string_spec(), formatted_string())
    )

mapping_spec = lambda: sb.create_spec(Mapping
    , content_type = sb.defaulted(formatted_string(), "application/json")
    , template = sb.defaulted(formatted_dictionary_or_string(), "$input.json('$')")
    )

class aws_resource_spec(Spec):
    def setup(self, method, resource_name):
        self.method = method
        self.resource_name = resource_name

    def normalise(self, meta, val):
        result = sb.create_spec(LambdaMethod
            , http_method = sb.overridden(self.method)
            , resource_name = sb.overridden(self.resource_name)

            , function = formatted_string()
            , location = formatted_string()
            , account = sb.optional_spec(formatted_string())
            , require_api_key = sb.defaulted(sb.boolean(), False)
            , request_mapping = sb.defaulted(mapping_spec(), Mapping("application/json", ""))
            , mapping = sb.defaulted(mapping_spec(), Mapping("application/json", "$input.json('$')"))
            , sample_event = sb.or_spec(formatted_dictionary(), sb.string_spec())
            , desired_output_for_test = sb.or_spec(formatted_dictionary(), sb.string_spec())
            ).normalise(meta, val)

        for key in ('sample_event', 'desired_output_for_test'):
            if isinstance(result[key], six.string_types):
                v = result[key]
                if v.startswith("{") and v.endswith("}"):
                    v = sb.formatted(sb.string_spec(), formatter=MergedOptionStringFormatter).normalise(meta.at(key), v)
                    result[key] = v

        function = result.function
        location = None

        if result.location is not NotSpecified and location is not None:
            raise BadSpecValue("Please don't specify a defined lambda function and location at the same time", meta=meta)

        if not isinstance(function, six.string_types):
            location = function.location
            function = function.name

        if location is None and result.location is NotSpecified:
            raise BadSpecValue("Location is a required key!", meta=meta)

        result.function = function
        result.location = location
        return result

class mock_resource_spec(Spec):
    def setup(self, method, resource_name):
        self.method = method
        self.resource_name = resource_name

    def normalise(self, meta, val):
        return sb.create_spec(MockMethod
            , http_method = sb.overridden(self.method)
            , resource_name = sb.overridden(self.resource_name)

            , request_mapping = sb.defaulted(mapping_spec(), Mapping("application/json", '{"statusCode": 200}'))
            , mapping = mapping_spec()
            , require_api_key = sb.defaulted(sb.boolean(), False)
            , sample_event = sb.or_spec(sb.dictionary_spec(), sb.string_spec())
            , desired_output_for_test = sb.or_spec(sb.dictionary_spec(), sb.string_spec())
            ).normalise(meta, val)

        for key in ('sample_event', 'desired_output_for_test'):
            if isinstance(result[key], six.string_types):
                v = result[key]
                if v.startswith("{") and v.endswith("}"):
                    v = sb.formatted(sb.string_spec(), formatter=MergedOptionStringFormatter).normalise(meta.at(key), v)
                    result[key] = v

class gateway_methods_spec(Spec):
    def normalise(self, meta, val):
        # Make sure we have integration
        integration_spec = sb.required(sb.string_choice_spec(["aws", "mock"]))
        sb.set_options(integration=integration_spec).normalise(meta, val)

        # Determine the http method and resource name
        method = meta.key_names()["_key_name_0"]
        resource_name = meta.key_names()["_key_name_2"]

        # We have integration if no exception was raised
        if val['integration'] == "aws":
            return aws_resource_spec(method, resource_name).normalise(meta, val)
        else:
            return mock_resource_spec(method, resource_name).normalise(meta, val)

gateway_resource_spec = lambda: sb.create_spec(GatewayResource
    , methods = sb.dictof(sb.string_spec(), gateway_methods_spec())
    )

class ApiKey(dictobj):
    fields = ['name', 'stages']

class DomainName(dictobj):
    fields = ['name', 'zone', 'stage', 'base_path', 'certificate', 'gateway_location']

    @property
    def full_name(self):
        return "{0}.{1}".format(self.name, self.zone)

    def cname(self, amazon):
        return amazon.apigateway.cname_for(self.gateway_location, self.full_name)

class Mapping(dictobj):
    fields = ['content_type', 'template']

class ResourceOptions(dictobj):
    fields = ['method_request', 'integration_request', 'method_response', 'integration_response']

class MethodExecutionRequest(dictobj):
    fields = ['require_api_key']

class MethodExecutionIntegrationRequest(dictobj):
    fields = ['integration_type', ('options', None)]

    def put_kwargs(self, gateway_location, accounts, environment):
        if self.options is None:
            kwargs = {}
        else:
            kwargs = self.options.put_kwargs(gateway_location, accounts, environment)
        kwargs['type'] = self.integration_type
        return kwargs

    def create_permissions(self, amazon, gateway_arn, gateway_name, accounts, environment):
        if self.integration_type == "AWS":
            self.options.create_permissions(amazon, gateway_arn, gateway_name, accounts, environment)

    def announce_create_permissions(self, gateway_name, changer):
        if self.integration_type == "AWS":
            self.options.announce_create_permissions(gateway_name, changer)

class MethodExecutionResponse(dictobj):
    fields = ['responses']

class MethodExecutionIntegrationResponse(dictobj):
    fields = ['responses']

class LambdaIntegrationOptions(dictobj):
    fields = ['http_method', 'resource_name', 'function', 'location', 'account', 'mapping']

    def arn(self, accounts, environment):
        if self.account is NotSpecified:
            account = accounts[environment]
        else:
            if self.account in accounts:
                account = accounts[self.account]
            else:
                account = self.account

        return "arn:aws:lambda:{0}:{1}:function:{2}".format(self.location, account, self.function)

    def put_kwargs(self, gateway_location, accounts, environment):
        arn = self.arn(accounts, environment)
        uri = "arn:aws:apigateway:{0}:lambda:path/2015-03-31/functions/{1}/invocations".format(gateway_location, arn)
        template = self.mapping.template
        if not isinstance(template, six.string_types):
            template = json.dumps(template, sort_keys=True)
        request_templates = {self.mapping.content_type: template}
        return {'uri': uri, 'requestTemplates': request_templates, 'httpMethod': "POST"}

    def create_permissions(self, amazon, gateway_arn, gateway_name, accounts, environment):
        arn = self.arn(accounts, environment)
        gateway_arn = "{0}{1}{2}".format(gateway_arn, self.http_method, self.resource_name)
        amazon.lambdas.modify_resource_policy_for_gateway(arn, self.location, gateway_arn, gateway_name)

    def announce_create_permissions(self, gateway_name, changer):
        # Purely for announcing the change we want to make
        for _ in changer("M", "Lambda resource policy", gateway=gateway_name, function=self.function):
            pass

class MockIntegrationOptions(dictobj):
    fields = ['http_method', 'mapping']

    def put_kwargs(self, gateway_location, accounts, environment):
        template = self.mapping.template
        if not isinstance(template, six.string_types):
            template = json.dumps(template, sort_keys=True)
        return {'requestTemplates': {self.mapping.content_type: template}, 'httpMethod': self.http_method}

class LambdaMethod(dictobj):
    fields = ['http_method', 'resource_name', 'function', 'location', 'account', 'require_api_key', 'request_mapping', 'mapping', 'sample_event', 'desired_output_for_test']

    @property
    def resource_options(self):
        return ResourceOptions(
              method_request = MethodExecutionRequest(require_api_key=self.require_api_key)
            , integration_request = MethodExecutionIntegrationRequest(integration_type="AWS", options=LambdaIntegrationOptions(mapping=self.request_mapping, http_method=self.http_method, resource_name=self.resource_name, function=self.function, location=self.location, account=self.account))
            , method_response = MethodExecutionResponse(responses={200: self.mapping.content_type})
            , integration_response = MethodExecutionIntegrationResponse(responses={200: [self.mapping]})
            )

class MockMethod(dictobj):
    fields = ['http_method', 'resource_name', 'request_mapping', 'mapping', 'require_api_key', 'sample_event', 'desired_output_for_test']

    @property
    def resource_options(self):
        return ResourceOptions(
              method_request = MethodExecutionRequest(require_api_key=self.require_api_key)
            , integration_request = MethodExecutionIntegrationRequest(integration_type="MOCK", options=MockIntegrationOptions(http_method=self.http_method, mapping=self.request_mapping))
            , method_response = MethodExecutionResponse(responses={200: "application/json"})
            , integration_response = MethodExecutionIntegrationResponse(responses={200: [self.mapping]})
            )

class GatewayResource(dictobj):
    fields = ['methods']

    @property
    def method_options(self):
        for key, val in self.methods.items():
            yield val.http_method, val.resource_options

class gateways_spec(Spec):
    def normalise(self, meta, val):
        if 'use' in val:
            template = val['use']
            if template not in meta.everything['templates']:
                available = list(meta.everything['templates'].keys())
                raise BadTemplate("Template doesn't exist!", wanted=template, available=available, meta=meta)

            val = MergedOptions.using(meta.everything['templates'][template], val)

        gateway_name = meta.key_names()['_key_name_0']
        gateway_location = formatted_string().normalise(meta.at('location'), val.get('location', ''))

        return sb.create_spec(Gateway
            , name = sb.overridden(gateway_name)
            , location = sb.required(formatted_string())
            , stages = sb.listof(formatted_string())
            , api_keys = sb.listof(api_key_spec())
            , domain_names = sb.dictof(sb.string_spec(), custom_domain_name_spec(gateway_location))
            , resources = sb.dictof(sb.string_spec(), gateway_resource_spec())
            ).normalise(meta, val)

class Secret(dictobj):
    fields = ['plain', 'kms', 'location', 'kms_data_key']

    def resolve(self, amazon):
        if self.plain is not NotSpecified:
            return self.plain
        else:
            data_key = amazon.kms.decrypt(self.location, self.kms_data_key)
            counter = Counter.new(128)
            decryptor = AES.new(data_key[:32], AES.MODE_CTR, counter=counter)
            return decryptor.decrypt(base64.b64decode(self.kms)).decode('utf-8')

class Certificate(dictobj):
    fields = ['name', 'body', 'key', 'chain']

class Gateways(dictobj):
    fields = ['items']

    def sync_one(self, aws_syncr, amazon, gateway):
        """Make sure this gateway exists and has only attributes we want it to have"""
        gateway_info = amazon.apigateway.gateway_info(gateway.name, gateway.location)
        if not gateway_info:
            amazon.apigateway.create_gateway(gateway.name, gateway.location, gateway.stages, gateway.resources, gateway.api_keys, gateway.domain_names)
        else:
            amazon.apigateway.modify_gateway(gateway_info, gateway.name, gateway.location, gateway.stages, gateway.resources, gateway.api_keys, gateway.domain_names)

class Gateway(dictobj):
    fields = {
          'name': "Name of the gateway"
        , 'location': "The region the gateway exists in"
        , 'stages': "The deployment stages for the gateway"
        , 'resources': "The resources in the gateway"
        , "api_keys": "The api keys to associate with this gateway"
        , "domain_names": "The custom domain names to associate with the gateway"
        }

    @property
    def stage_names(self):
        return list(self.stages)

    def gateway_info(self, amazon):
        if not getattr(self, "_gateway_info", None):
            self._gateway_info = amazon.apigateway.gateway_info(self.name, self.location)
        return self._gateway_info

    def validate_stage(self, amazon, stage):
        if stage not in self.stage_names:
            raise UnknownStage("Please specify a defined stage", available=self.stage_names)

        log.info("Finding information for gateway {0}".format(self.name))
        if not self.gateway_info(amazon):
            raise UnsyncedGateway("Please do a sync before trying to deploy your gateway!")

        defined_stages = [stage['stageName'] for stage in self.gateway_info(amazon)['stages']]
        if stage not in defined_stages:
            raise UnknownStage("Please do a sync before trying to deploy your gateway!", only_have=defined_stages)

    def find_sample_event(self, amazon, method, endpoint):
        available = list(self.resources.keys())
        if endpoint not in available:
            raise UnknownEndpoint("Please specify an endpoint that exists", got=endpoint, available=available)

        methods = self.resources[endpoint].methods
        if method not in self.resources[endpoint].methods:
            raise UnknownEndpoint("Please specify a valid http_method for this endpoint", got=method, available=list(methods.keys()))

        return self.resources[endpoint].methods[method].sample_event, self.resources[endpoint].methods[method].desired_output_for_test

    def available_methods_and_endpoints(self):
        for endpoint, resource in self.resources.items():
            for method in resource.methods:
                yield method, endpoint

    def deploy(self, aws_syncr, amazon, stage):
        self.validate_stage(amazon, stage)
        amazon.apigateway.deploy_stage(self.gateway_info(amazon), self.location, stage, aws_syncr.extra)

    def test(self, aws_syncr, amazon, stage):
        endpoint = aws_syncr.extra.strip()
        if not endpoint or " " not in endpoint:
            options = sorted("{0} {1}".format(m, e) for m, e in list(self.available_methods_and_endpoints()))
            raise AwsSyncrError("{0}\n".format(dedent("""
            Please specify ' -- <http_method> <endpoint> ' at the end of the command

            For example:

                {0} -- <http_method> <endpoint>

            Where the available options are:

                {1}

            """.format(' '.join(sys.argv), '\n\t\t'.join(options)).strip()
            )))

        method, endpoint = endpoint.split(" ", 1)
        sample_event, desired_output_for_test = self.find_sample_event(amazon, method, endpoint)

        self.validate_stage(amazon, stage)
        return amazon.apigateway.test_stage(self.gateway_info(amazon), self.location, stage, method, endpoint, sample_event, desired_output_for_test)

def __register__():
    return {(99, "apigateway"): sb.container_spec(Gateways, sb.dictof(sb.string_spec(), gateways_spec()))}




#!/usr/bin/env python
"""
This is where the mainline sits and is responsible for setting up the logging,
the argument parsing and for starting up aws_syncr.
"""

from aws_syncr.actions import available_actions
from aws_syncr.collector import Collector
from aws_syncr.errors import BadTask

from delfick_app import App
import logging

log = logging.getLogger("aws_syncr.executor")

class App(App):
    cli_categories = ['aws_syncr']
    cli_description = "Application that reads YAML and syncs definitions with amazon"
    cli_environment_defaults = {"AWS_SYNCR_CONFIG_FOLDER": ("--config-folder", '.')}
    cli_positional_replacements = [('--environment'), ('--task', 'list_tasks'), ('--artifact', "")]

    def execute(self, args, extra_args, cli_args, logging_handler, no_docker=False):
        cli_args["aws_syncr"]["extra"] = extra_args
        cli_args["aws_syncr"]["debug"] = args.debug

        collector = Collector()
        collector.prepare(cli_args["aws_syncr"]["config_folder"], cli_args, cli_args['aws_syncr']['environment'])
        if hasattr(collector, "configuration") and "term_colors" in collector.configuration:
            self.setup_logging_theme(logging_handler, colors=collector.configuration["term_colors"])

        task = args.aws_syncr_chosen_task
        if task not in available_actions:
            raise BadTask("Unknown task", available=list(available_actions.keys()), wanted=task)

        available_actions[task](collector)

    def setup_other_logging(self, args, verbose=False, silent=False, debug=False):
        logging.getLogger("boto3").setLevel([logging.CRITICAL, logging.DEBUG][verbose or debug])
        logging.getLogger("requests").setLevel([logging.CRITICAL, logging.ERROR][verbose or debug])
        logging.getLogger("botocore").setLevel([logging.CRITICAL, logging.DEBUG][verbose or debug])

    def specify_other_args(self, parser, defaults):
        parser.add_argument("--config-folder"
            , help = "The config folder containing the environments aws_syncr should care about"
            , dest = "aws_syncr_config_folder"
            , **defaults["--config-folder"]
            )

        parser.add_argument("--dry-run"
            , help = "Should aws_syncr take any real action or print out what is intends to do"
            , dest = "aws_syncr_dry_run"
            , action = "store_true"
            )

        parser.add_argument("--task"
            , help = "The task to run"
            , dest = "aws_syncr_chosen_task"
            , **defaults["--task"]
            )

        parser.add_argument("--environment"
            , help = "Environment to read options from"
            , dest = "aws_syncr_environment"
            , **defaults['--environment']
            )

        parser.add_argument("--artifact"
            , help = "Extra argument to be used as decided by each task"
            , dest = "aws_syncr_artifact"
            , **defaults['--artifact']
            )

        parser.add_argument("--stage"
            , help = "Extra argument to be used as the stage for deploying an api gateway"
            , dest = "aws_syncr_stage"
            , default = ""
            )

        return parser

main = App.main
if __name__ == '__main__':
    main()



from aws_syncr.formatter import MergedOptionStringFormatter
from aws_syncr.option_spec.resources import resource_spec
from aws_syncr.errors import BadTemplate

from input_algorithms.spec_base import NotSpecified
from input_algorithms.errors import BadSpecValue
from input_algorithms import spec_base as sb
from input_algorithms.dictobj import dictobj
from input_algorithms.spec_base import Spec
from option_merge import MergedOptions
from contextlib import contextmanager
from textwrap import dedent
import tempfile
import logging
import fnmatch
import zipfile
import json
import six
import re
import os

log = logging.getLogger("aws_syncr.option_spec.lambdas")

class formatted_dictionary(sb.Spec):
    def normalise(self, meta, val):
        val = sb.dictionary_spec().normalise(meta, val)
        return self.formatted_dict(meta, val)

    def formatted_dict(self, meta, val, chain=None):
        result = {}
        for key, val in val.items():
            if type(val) is dict:
                result[key] = self.formatted_dict(meta.at(key), val, chain)
            elif isinstance(val, six.string_types):
                result[key] = sb.formatted(sb.string_spec(), formatter=MergedOptionStringFormatter).normalise(meta.at(key), val)
            else:
                result[key] = val
        return result

class only_one_spec(sb.Spec):
    def setup(self, spec):
        self.spec = spec

    def normalise(self, meta, val):
        val = self.spec.normalise(meta, val)
        if type(val) is not list:
            return val

        if len(val) != 1:
            raise BadSpecValue("Please only specify one value", meta=meta)

        return val[0]

class divisible_by_spec(sb.Spec):
    def setup(self, divider):
        self.divider = divider

    def normalise_filled(self, meta, val):
        val = sb.integer_spec().normalise(meta, val)
        if val % self.divider != 0:
            raise BadSpecValue("Value should be divisible by {0}".format(self.divider), meta=meta)
        return val

class function_handler_spec(sb.Spec):
    def normalise_empty(self, meta):
        path = [p for p, _ in meta._path]
        path.pop()
        runtime = meta.everything['.'.join(path)].get("runtime", "python")
        runtime = sb.formatted(sb.string_spec(), formatter=MergedOptionStringFormatter).normalise(meta.at("runtime"), runtime)

        if runtime == 'java8':
            raise BadSpecValue("No default function handler for java", meta=meta)
        elif runtime == 'nodejs':
            return "index.handler"
        elif runtime == 'python2.7':
            return "lambda_function.lambda_handler"
        else:
            raise BadSpecValue("No default function handler for {0}".format(runtime), meta=meta)

    def normalise_filled(self, meta, val):
        return sb.formatted(sb.string_spec(), formatter=MergedOptionStringFormatter).normalise(meta, val)

class function_code_spec(sb.Spec):
    def normalise_filled(self, meta, val):
        val = sb.dictof(sb.string_choice_spec(["s3", "inline", "directory"]), sb.any_spec()).normalise(meta, val)
        if not val:
            raise BadSpecValue("Please specify s3, inline or directory for your code", meta=meta)

        if len(val) > 1:
            raise BadSpecValue("Please only specify one of s3, inline or directory for your code", got=list(val.keys()), meta=meta)

        formatted_string = sb.formatted(sb.string_spec(), formatter=MergedOptionStringFormatter)
        if "s3" in val:
            return sb.create_spec(S3Code
                , key = formatted_string
                , bucket = formatted_string
                , version = sb.defaulted(sb.string_spec(), NotSpecified)
                ).normalise(meta, val['s3'])
        elif "inline" in val:
            path = [p for p, _ in meta._path]
            path.pop()
            runtime = meta.everything['.'.join(path)].get("runtime", "python")
            runtime = sb.formatted(sb.string_spec(), formatter=MergedOptionStringFormatter).normalise(meta.at("runtime"), runtime)

            return sb.create_spec(InlineCode
                , code = sb.string_spec()
                , runtime = sb.overridden(runtime)
                ).normalise(meta, {"code": val['inline']})
        else:
            directory = val['directory']
            if isinstance(val['directory'], six.string_types):
                directory = {"directory": val['directory']}

            if 'directory' in directory:
                formatted_string = sb.formatted(sb.string_spec(), formatter=MergedOptionStringFormatter)
                directory['directory'] = formatted_string.normalise(meta.at("directory").at("directory"), directory['directory'])

            return sb.create_spec(DirectoryCode
                , directory = sb.directory_spec()
                , exclude = sb.listof(sb.string_spec())
                ).normalise(meta, directory)

class lambdas_spec(Spec):
    def normalise(self, meta, val):
        if 'use' in val:
            template = val['use']
            if template not in meta.everything['templates']:
                available = list(meta.everything['templates'].keys())
                raise BadTemplate("Template doesn't exist!", wanted=template, available=available, meta=meta)

            val = MergedOptions.using(meta.everything['templates'][template], val)

        formatted_string = sb.formatted(sb.string_or_int_as_string_spec(), MergedOptionStringFormatter, expected_type=six.string_types)
        function_name = meta.key_names()['_key_name_0']

        val = sb.create_spec(Lambda
            , name = sb.overridden(function_name)
            , role = sb.required(only_one_spec(resource_spec("lambda", function_name, only=["iam"])))
            , code = sb.required(function_code_spec())
            , handler = function_handler_spec()
            , timeout = sb.integer_spec()
            , runtime = sb.required(formatted_string)
            , location = sb.required(formatted_string)
            , description = formatted_string
            , sample_event = sb.defaulted(sb.or_spec(formatted_dictionary(), sb.string_spec()), "")
            , desired_output_for_test = sb.defaulted(sb.or_spec(formatted_dictionary(), sb.string_spec()), "")
            , memory_size = sb.defaulted(divisible_by_spec(64), 128)
            ).normalise(meta, val)

        # Hack to make sample_event and desired_output_for_test not appear as a MergedOptions
        for key in ('sample_event', 'desired_output_for_test'):
            if isinstance(val[key], MergedOptions):
                v = val[key].as_dict()
                class Arbritrary(dictobj):
                    fields = list(v.keys())
                val[key] = Arbritrary(**v)
        return val

class Lambdas(dictobj):
    fields = ['items']

    def sync_one(self, aws_syncr, amazon, function):
        """Make sure this function exists and has only attributes we want it to have"""
        function_info = amazon.lambdas.function_info(function.name, function.location)
        if not function_info:
            amazon.lambdas.create_function(function.name, function.description, function.location, function.runtime, function.role, function.handler, function.timeout, function.memory_size, function.code)
        else:
            amazon.lambdas.modify_function(function_info, function.name, function.description, function.location, function.runtime, function.role, function.handler, function.timeout, function.memory_size, function.code)

class Lambda(dictobj):
    fields = {
          'name': "Alias of the function"
        , 'role': "The role assumed by the function"
        , 'code': "Code for the function!"
        , 'handler': "Function within your code that gets executed"
        , 'timeout': "Max function execution time"
        , 'runtime': "Runtime environment for the function"
        , 'location': "The region the function exists in"
        , 'description': "Description of the function"
        , 'sample_event': "A sample event to test with"
        , 'desired_output_for_test': "Keys and values for the output of the test to consider it successful"
        , 'memory_size': "Max memory size for the function"
        }

    def deploy(self, aws_syncr, amazon):
        print(json.dumps(amazon.lambdas.deploy_function(self.name, self.code, self.location), indent=4))

    def test(self, aws_syncr, amazon):
        output = amazon.lambdas.test_function(self.name, self.sample_event, self.location)
        print(json.dumps(output, indent=4))

        if self.desired_output_for_test and self.desired_output_for_test is not NotSpecified:
            content = output['Payload']
            if isinstance(self.desired_output_for_test, six.string_types):
                if not re.match(self.desired_output_for_test, content):
                    print("content '{0}' does not match pattern '{1}'".format(content, self.desired_output_for_test))
                    return False

            else:
                if any(key not in content or content[key] != val for key, val in self.desired_output_for_test.items()):
                    print("Not all of the values match our desired output of '{0}'".format(self.desired_output_for_test))
                    return False

        return True

class S3Code(dictobj):
    fields = ["key", "bucket", "version"]

    @property
    def s3_address(self):
        return "s3://{0}/{1}".format(self.bucket, self.key)

    @contextmanager
    def zipfile(self):
        yield

class InlineCode(dictobj):
    fields = ["code", "runtime"]
    s3_address = None

    @property
    def arcname(self):
        if self.runtime == "python2.7":
            return "./lambda_function.py"
        elif self.runtime == "java8":
            return "./main.java"
        else:
            return "./index.js"

    @contextmanager
    def code_in_file(self):
        with tempfile.NamedTemporaryFile() as fle:
            fle.write(dedent(self.code).encode('utf-8'))
            fle.flush()
            yield fle.name

    @contextmanager
    def zipfile(self):
        with tempfile.NamedTemporaryFile(suffix=".zip") as fle:
            with self.code_in_file() as filename:
                log.info("Making zipfile")
                with zipfile.ZipFile(fle.name, "w") as zf:
                    zf.write(filename, self.arcname)
            yield fle.name

class DirectoryCode(dictobj):
    fields = ["directory", "exclude"]
    s3_address = None

    def files(self):
        for root, dirs, files in os.walk(self.directory, followlinks=True):
            for fle in files:
                location = os.path.join(root, fle)
                if not any(fnmatch.fnmatch(location, os.path.join(self.directory, ex)) for ex in self.exclude):
                    yield location, os.path.relpath(location, self.directory)

    @contextmanager
    def zipfile(self):
        with tempfile.NamedTemporaryFile(suffix=".zip") as fle:
            log.info("Making zipfile")
            with zipfile.ZipFile(fle.name, "w") as zf:
                for filename, arcname in self.files():
                    zf.write(filename, arcname)
            yield fle.name

def __register__():
    return {(22, "lambda"): sb.container_spec(Lambdas, sb.dictof(sb.string_spec(), lambdas_spec()))}




from aws_syncr.formatter import MergedOptionStringFormatter
from aws_syncr.errors import BadPolicy

from input_algorithms.spec_base import NotSpecified
from input_algorithms import spec_base as sb
import six

class resource_spec_base(sb.Spec):
    def setup(self, resource, self_type, self_name):
        self.resource = resource
        self.self_type = self_type
        self.self_name = self_name

    def default_account_id(self, meta):
        accounts = meta.everything["accounts"]
        return accounts[meta.everything["aws_syncr"].environment]

    def accounts(self, meta):
        accounts = meta.everything["accounts"]
        default_account_id = self.default_account_id(meta)

        provided_accounts = sb.listof(sb.string_spec()).normalise(meta.at("account"), self.resource.get("account", []))
        if not provided_accounts:
            yield default_account_id

        for provided_account in provided_accounts:
            if not provided_account:
                yield ""
            else:
                if provided_account not in accounts:
                    raise BadPolicy("Unknown account specified", account=provided_account, meta=meta)
                else:
                    account_id = accounts[provided_account]

                yield account_id

    def default_location(self, meta):
        return meta.everything["aws_syncr"].location

    def location(self, meta):
        return sb.defaulted(sb.string_spec(), self.default_location(meta)).normalise(meta.at("location"), self.resource.get("location", NotSpecified))

class iam_specs(resource_spec_base):
    def normalise(self, meta, val):
        pairs = []
        has_self = False
        for account_id in self.accounts(meta):
            users = sb.listof(sb.string_spec()).normalise(meta.at("users"), self.resource.get('users', NotSpecified))
            for index, name in enumerate(sb.listof(sb.any_spec()).normalise(meta, val)):
                if name == "__self__":
                    if self.self_type != 'role':
                        raise BadPolicy("No __self__ iam role for this policy", meta=meta)
                    else:
                        has_self = True
                else:
                    if isinstance(name, six.string_types):
                        name = sb.formatted(sb.string_spec(), formatter=MergedOptionStringFormatter).normalise(meta.indexed_at(index), name)
                    pairs.append((name, account_id))

        if has_self:
            pairs.append(("role/{0}".format(self.self_name), self.default_account_id(meta)))

        for name, account_id in pairs:
            service = "sts" if name.startswith("assumed-role") else "iam"
            arn = "arn:aws:{0}::{1}:{2}".format(service, account_id, name)
            if not users:
                yield arn
            else:
                for user in users:
                    yield "{0}/{1}".format(arn, user)

class s3_specs(resource_spec_base):
    def normalise(self, meta, val):
        for bucket_key in sb.listof(sb.formatted(sb.string_spec(), formatter=MergedOptionStringFormatter)).normalise(meta, val):
            if bucket_key == "__self__" or bucket_key.startswith("__self__"):
                if self.self_type != "bucket":
                    raise BadPolicy("No __self__ bucket for this policy", meta=meta)
                else:
                    path = ""
                    if "/" in bucket_key:
                        path = bucket_key[bucket_key.find('/'):]
                    bucket_key = "{0}{1}".format(self.self_name, path)

            yield "arn:aws:s3:::{0}".format(bucket_key)
            if '/' not in bucket_key:
                yield "arn:aws:s3:::{0}/*".format(bucket_key)

class kms_specs(resource_spec_base):
    def normalise(self, meta, val):
        accounts = list(self.accounts(meta))
        if not accounts:
            accounts = [self.default_account_id(meta)]

        for account_id in accounts:
            string_or_dict = sb.or_spec(sb.string_spec(), sb.dictof(sb.string_choice_spec(["key_id", "alias"]), sb.string_spec()))
            for key_id in sb.listof(string_or_dict).normalise(meta, val):
                alias = None
                if key_id == "__self__" or (isinstance(key_id, dict) and (key_id.get("alias") == "__self__" or key_id.get("key_id") == "__self__")):
                    if self.self_type != "key":
                        raise BadPolicy("No __self__ key for this policy", meta=meta)
                    else:
                        alias = self.self_name
                        location = self.default_location(meta)
                else:
                    location = self.location(meta)

                if not alias:
                    if isinstance(key_id, six.string_types):
                        alias = key_id
                    else:
                        alias = key_id.get("alias")
                        key_id = key_id.get("key_id")

                if alias:
                    yield "arn:aws:kms:{0}:{1}:alias/{2}".format(location, account_id, alias)
                else:
                    yield "arn:aws:kms:{0}:{1}:key/{2}".format(location, account_id, key_id)

class arn_specs(resource_spec_base):
    def normalise(self, meta, val):
        default_location = ""

        if "identity" not in self.resource:
            raise BadPolicy("Generic arn specified without specifying 'identity'", meta=meta)

        location = self.location(meta)
        identities = sb.listof(sb.formatted(sb.string_spec(), formatter=MergedOptionStringFormatter)).normalise(meta.at("identity"), self.resource.get("identity"))

        for account_id in self.accounts(meta):
            for identity in identities:
                yield "arn:aws:{0}:{1}:{2}:{3}".format(val, location, account_id, identity)

class resource_spec(sb.Spec):
    def setup(self, self_type, self_name, only=None):
        self.only = only
        self.self_type = self_type
        self.self_name = self_name

    def normalise(self, meta, val):
        result = []
        for index, item in enumerate(sb.listof(sb.any_spec()).normalise(meta, val)):
            s3_spec = s3_specs(item, self.self_type, self.self_name)
            iam_spec = iam_specs(item, self.self_type, self.self_name)
            kms_spec = kms_specs(item, self.self_type, self.self_name)
            arn_spec = arn_specs(item, self.self_type, self.self_name)

            if isinstance(item, six.string_types):
                result.append(item)
            else:
                types = (("iam", iam_spec), ("kms", kms_spec), ("s3", s3_spec), ("arn", arn_spec))
                for typ, spec in types:
                    if typ in item:
                        if self.only and typ not in self.only:
                            raise BadPolicy("Sorry, don't support this resource type here", wanted=typ, available=self.only, meta=meta)

                        for found in spec.normalise(meta.indexed_at(index).at(typ), item[typ]):
                            result.append(found)
        return sorted(result)




from aws_syncr.errors import BadCredentials, AwsSyncrError
from aws_syncr.amazon.apigateway import ApiGateway
from aws_syncr.amazon.common import AmazonMixin
from aws_syncr.amazon.lambdas import Lambdas
from aws_syncr.amazon.route53 import Route53
from aws_syncr.amazon.iam import Iam
from aws_syncr.amazon.kms import Kms
from aws_syncr.amazon.s3 import S3
import boto3

import logging

log = logging.getLogger("aws_syncr.amazon.amazon")

class ValidatingMemoizedProperty(object):
    def __init__(self, kls, key):
        self.kls = kls
        self.key = key

    def __get__(self, instance, owner):
        obj = getattr(instance, self.key, None)
        if not obj:
            if not getattr(instance, "_validated", False) and not getattr(instance, "_validating", False):
                instance.validate_account()
            obj = self.kls(instance, instance.environment, instance.accounts, instance.dry_run)
            setattr(instance, self.key, obj)
        return obj

class Amazon(AmazonMixin, object):
    def __init__(self, environment, accounts, debug=False, dry_run=False):
        self.debug = debug
        self.dry_run = dry_run
        self.accounts = accounts
        self.environment = environment

        self.changes = False
        self.session = boto3.session.Session()

    s3 = ValidatingMemoizedProperty(S3, "_s3")
    iam = ValidatingMemoizedProperty(Iam, "_iam")
    kms = ValidatingMemoizedProperty(Kms, "_kms")
    lambdas = ValidatingMemoizedProperty(Lambdas, "_lambdas")
    route53 = ValidatingMemoizedProperty(Route53, "_route53")
    apigateway = ValidatingMemoizedProperty(ApiGateway, "_apigateway")

    def validate_account(self):
        """Make sure we are able to connect to the right account"""
        self._validating = True
        with self.catch_invalid_credentials():
            log.info("Finding a role to check the account id")
            a_role = list(self.iam.resource.roles.limit(1))
            if not a_role:
                raise AwsSyncrError("Couldn't find an iam role, can't validate the account....")
            account_id = a_role[0].meta.data['Arn'].split(":", 5)[4]

        chosen_account = self.accounts[self.environment]
        if chosen_account != account_id:
            raise BadCredentials("Don't have credentials for the correct account!", wanted=chosen_account, got=account_id)

        self._validating = False
        self._validated = True




from aws_syncr.amazon.common import AmazonMixin
from aws_syncr.differ import Differ

import logging
import base64

log = logging.getLogger("aws_syncr.amazon.kms")

class Kms(AmazonMixin, object):
    def __init__(self, amazon, environment, accounts, dry_run):
        self.amazon = amazon
        self.dry_run = dry_run

        self.accounts = accounts
        self.account_id = accounts[environment]
        self.environment = environment

        self.clients = {}

    def get_client(self, location):
        if location not in self.clients:
            self.clients[location] = self.amazon.session.client('kms', location)
        return self.clients[location]

    def decrypt(self, location, secret):
        return self.get_client(location).decrypt(CiphertextBlob=base64.b64decode(secret))['Plaintext']

    def generate_data_key(self, location, key_id):
        return self.get_client(location).generate_data_key(KeyId=key_id, KeySpec="AES_256")

    def key_info(self, name, location):
        client = self.get_client(location)
        response = None
        with self.ignore_missing():
            response = client.describe_key(KeyId="alias/{0}".format(name))

        if response is None:
            return {}

        info = {"KeyId": response["KeyMetadata"]["KeyId"], "Description": response["KeyMetadata"]["Description"]}
        info['Policy'] = client.get_key_policy(KeyId=info["KeyId"], PolicyName="default")["Policy"]
        info['Grants'] = client.list_grants(KeyId=info["KeyId"])["Grants"]

        return info

    def create_key(self, name, description, location, grant, policy):
        client = self.get_client(location)
        with self.catch_boto_400("Couldn't create key", "{0} Policy".format(name), policy, alias=name):
            for _ in self.change("+", "kms_key", alias=name, document=policy):
                keyid = client.create_key(Description=description, Policy=policy)["KeyMetadata"]["KeyId"]

                with self.catch_boto_400("Couldn't create alias", alias=name, keyid=keyid):
                    client.create_alias(AliasName="alias/{0}".format(name), TargetKeyId=keyid)

                self.handle_grants(client, keyid, [], name, [g.statement for g in grant])

    def modify_key(self, key_info, name, description, location, grant, policy):
        client = self.get_client(location)
        if key_info["Description"] != description:
            with self.catch_boto_400("Couldn't change the description", alias=name):
                for _ in self.change("M", "kms_description", alias=name):
                    client.update_key_description(KeyId=key_info["KeyId"], Description=description)

        changes = list(Differ.compare_two_documents(key_info["Policy"], policy))
        if changes:
            with self.catch_boto_400("Couldn't modify policy", "Key {0} policy".format(name), policy, bucket=name):
                for _ in self.change("M", "key_policy", alias=name, changes=changes):
                    client.put_key_policy(KeyId=key_info["KeyId"], Policy=policy, PolicyName="default")

        self.handle_grants(client, key_info["KeyId"], key_info["Grants"], name, [g.statement for g in grant])

    def handle_grants(self, client, keyid, current_grants, name, new_grants):
        new = []
        revokable = []
        NotFound = type("NotFound", (object, ), {})

        def match(lst, grant):
            for l in lst:
                match = True
                for k, v in grant.items():
                    if k not in ("GrantId", "IssuingAccount"):
                        if type(v) is list:
                            if sorted(v) != sorted(l.get(k, [NotFound])):
                                match = False
                        elif k in l and l.get(k) != v:
                            match = False
                if match:
                    return True

        for grant in new_grants:
            if not match(current_grants, grant):
                new.append(grant)

        for grant in current_grants:
            if not match(new_grants, grant):
                revokable.append(grant)

        for grant in new:
            with self.catch_boto_400("Couldn't create grant", key=name):
                for _ in self.change("+", "key_grant", grantee=grant.get("GranteePrincipal"), retiree=grant.get("RetireePrincipal"), key=name):
                    client.create_grant(KeyId=keyid, **grant)

        for grant in revokable:
            with self.catch_boto_400("Couldn't revoke grant", key=name, grant=grant["GrantId"]):
                for _ in self.change("-", "key_grant", grantee=grant.get("GranteePrincipal"), retiree=grant.get("RetireePrincipal"), key=name, grant=grant["GrantId"]):
                    client.revoke_grant(KeyId=keyid, GrantId=grant["GrantId"])



from aws_syncr.amazon.common import AmazonMixin
from aws_syncr.errors import AwsSyncrError
from aws_syncr.differ import Differ

import logging
import json

log = logging.getLogger("aws_syncr.amazon.s3")

class S3(AmazonMixin, object):
    def __init__(self, amazon, environment, accounts, dry_run):
        self.amazon = amazon
        self.dry_run = dry_run

        self.accounts = accounts
        self.account_id = accounts[environment]
        self.environment = environment

        self.client = self.amazon.session.client("s3")
        self.resource = self.amazon.session.resource('s3')

    def bucket_info(self, bucket_name):
        bucket = self.resource.Bucket(bucket_name.split('/')[-1])
        with self.ignore_missing():
            bucket.load()
            return bucket

    def create_bucket(self, name, permission_document, location, tags):
        with self.catch_boto_400("Couldn't Make bucket", bucket=name):
            for _ in self.change("+", "bucket", bucket=name):
                self.resource.create_bucket(Bucket=name, CreateBucketConfiguration={"LocationConstraint": location})

        if permission_document:
            with self.catch_boto_400("Couldn't add policy", "{0} Permission document".format(name), permission_document, bucket=name):
                for _ in self.change("+", "bucket_policy", bucket=name, document=permission_document):
                    self.resource.Bucket(name).Policy().put(Policy=permission_document)

        if tags:
            with self.catch_boto_400("Couldn't add tags", bucket=name):
                tag_set = [{"Value": val, "Key": key} for key, val in tags.items()]
                changes = list(Differ.compare_two_documents("[]", json.dumps(tag_set)))
                for _ in self.change("+", "bucket_tags", bucket=name, tags=tags, changes=changes):
                    self.resource.Bucket(name).Tagging().put(Tagging={"TagSet": tag_set})

    def modify_bucket(self, bucket_info, name, permission_document, location, tags):
        current_location = self.client.get_bucket_location(Bucket=name)['LocationConstraint']
        if current_location != location:
            raise AwsSyncrError("Sorry, can't change the location of a bucket!", wanted=location, currently=current_location, bucket=name)

        # Make sure we use the correct endpoint to get info from the bucket
        # So that website buckets don't complain
        bucket_info.meta.client = self.amazon.session.client("s3", location)

        bucket_document = ""
        with self.ignore_missing():
            bucket_document = bucket_info.Policy().policy

        if bucket_document or permission_document:
            if permission_document and not bucket_document:
                with self.catch_boto_400("Couldn't add policy", "Bucket {0} policy".format(name), permission_document, bucket=name):
                    for _ in self.change("+", "bucket_policy", bucket=name, changes=list(Differ.compare_two_documents("{}", permission_document))):
                        bucket_info.Policy().put(Policy=permission_document)

            elif bucket_document and not permission_document:
                with self.catch_boto_400("Couldn't remove policy", "Bucket {0} policy".format(name), permission_document, bucket=name):
                    for _ in self.change("-", "bucket_policy", bucket=name, changes=list(Differ.compare_two_documents(bucket_document, "{}"))):
                        bucket_info.Policy().delete()

            else:
                changes = list(Differ.compare_two_documents(bucket_document, permission_document))
                if changes:
                    with self.catch_boto_400("Couldn't modify policy", "Bucket {0} policy".format(name), permission_document, bucket=name):
                        for _ in self.change("M", "bucket_policy", bucket=name, changes=changes):
                            bucket_info.Policy().put(Policy=permission_document)

        self.modify_tags(bucket_info, name, tags)

    def modify_tags(self, bucket_info, name, tags):
        current_tags = bucket_info.Tagging()
        tag_set = []
        with self.ignore_missing():
            current_tags.load()
            tag_set = current_tags.tag_set
        current_vals = dict((tag["Key"], tag["Value"]) for tag in tag_set)

        changes = list(Differ.compare_two_documents(json.dumps(current_vals), json.dumps(tags)))
        if changes:
            new_tag_set = [{"Value": val, "Key": key} for key, val in tags.items()]
            with self.catch_boto_400("Couldn't modify tags", bucket=name):
                symbol = "+" if new_tag_set else "-"
                symbol = "M" if new_tag_set and current_vals else symbol
                for _ in self.change(symbol, "bucket_tags", bucket=name, changes=changes):
                    if not new_tag_set:
                        bucket_info.Tagging().delete()
                    else:
                        bucket_info.Tagging().put(Tagging={"TagSet": new_tag_set})

